from typing import List, Optional
import click

from aws_iam_login.observer import Observer, Subject


class ApplicationMessages(Subject):
    __observers: List[Observer] = []
    message: str
    type: str
    exception: Optional[Exception]

    def attach(self, observer: Observer) -> None:
        self.__observers.append(observer)

    def detach(self, observer: Observer) -> None:
        self.__observers.remove(observer)

    def notify(self) -> None:
        for observer in self.__observers:
            observer.update(self)

    def send(
        self, message_type: str, message: str, exception: Optional[Exception] = None
    ) -> None:
        self.message = message
        self.type = message_type
        self.exception = exception
        self.notify()


class CliObserver(Observer):
    def __init__(self, debug: bool) -> None:
        self.__debug = debug

    def update(self, subject: Subject) -> None:
        if not isinstance(subject, ApplicationMessages):
            return

        click.echo(f"[{subject.type.upper()}] {subject.message}")

        if subject.exception and self.__debug:
            click.echo(f"Exception: {subject.exception}")


class ApplicationContext:
    def __init__(self, profile: str, debug: bool) -> None:
        self.__subject = ApplicationMessages()
        self.__subject.attach(CliObserver(debug=debug))
        self.profile = profile

    @property
    def subject(self) -> ApplicationMessages:
        return self.__subject



from __future__ import annotations
from abc import ABC, abstractmethod


class Subject(ABC):
    @abstractmethod
    def attach(self, observer: Observer) -> None:
        pass

    @abstractmethod
    def detach(self, observer: Observer) -> None:
        pass

    @abstractmethod
    def notify(self) -> None:
        pass


class Observer(ABC):
    @abstractmethod
    def update(self, subject: Subject) -> None:
        pass



from typing import Optional
from abc import ABC, abstractmethod
from aws_iam_login.application_context import ApplicationMessages


class Action(ABC):
    """
    Understands generic actions that are executed
    """

    __subject: Optional[ApplicationMessages] = None

    def subscribe_subject(self, subject: ApplicationMessages):
        self.__subject = subject

    def info(self, message: str) -> None:
        if self.__subject:
            self.__subject.send(message_type="INFO", message=message)

    def warning(self, message: str) -> None:
        if self.__subject:
            self.__subject.send(message_type="WARN", message=message)

    def error(self, message: str) -> None:
        if self.__subject:
            self.__subject.send(message_type="ERROR", message=message)

    def exception(self, message: str, exception: Exception) -> None:
        if self.__subject:
            self.__subject.send(
                message_type="ERROR", message=message, exception=exception
            )

    @abstractmethod
    def execute(self) -> bool:
        pass



from typing import Optional, Union
import os
import configparser

from aws_iam_login.credentials import Credentials
from aws_iam_login.access_key import AccessKey


class AWSConfig:
    __cached_config: Optional[configparser.ConfigParser]
    __cached_key: Optional[AccessKey]

    def __init__(self, profile: str) -> None:
        self.__cached_config = None
        self.__cached_key = None
        self.__credential_file = os.environ.get(
            "AWS_SHARED_CREDENTIALS_FILE", os.path.expanduser("~/.aws/credentials")
        )
        self.__profile_name = profile

    @property
    def valid(self) -> bool:
        return bool(
            self.__profile_name
            and self.__profile_name in self.__config
            and self.mfa_serial
            and self.username
        )

    @property
    def __config(self) -> configparser.ConfigParser:
        if not self.__cached_config:
            self.__cached_config = configparser.ConfigParser()
            self.__cached_config.read(self.__credential_file)

        return self.__cached_config

    @property
    def __profile(self) -> configparser.SectionProxy:
        return self.__config[self.__profile_name]

    @property
    def mfa_serial(self) -> str:
        return self.__profile.get("mfa_serial", "")

    @property
    def username(self) -> str:
        return self.__profile.get("username", "")

    @property
    def key(self) -> AccessKey:
        if not self.__cached_key:
            existing_profile = self.__config[self.__profile_name]
            self.__cached_key = AccessKey(
                {
                    "Profile": self.__profile_name,
                    "AccessKeyId": existing_profile["aws_access_key_id"],
                    "SecretAccessKey": existing_profile["aws_secret_access_key"],
                }
            )

        return self.__cached_key

    def initialize(self, username: str, mfa_serial: str) -> None:
        self.__profile["username"] = username
        self.__profile["mfa_serial"] = mfa_serial
        self.__save_configuration()

    def __save_configuration(self):
        with open(self.__credential_file, "w") as fh:
            self.__config.write(fh)

    def write(self, profile: str, credentials: Union[None, Credentials]) -> None:
        if not credentials:
            return

        for key, value in dict(credentials).items():
            if profile not in self.__config:
                self.__config[profile] = {}

            self.__config[profile][key] = value

        self.__save_configuration()



import click
import boto3
from botocore.exceptions import ParamValidationError, ClientError
from click import Context

from aws_iam_login.actions.initialize_configuration import InitializeConfiguration
from aws_iam_login.observer import Observer
from aws_iam_login.temp_credentials import TempCredentials
from aws_iam_login.aws_config import AWSConfig
from aws_iam_login.actions.rotate_access_keys import RotateAccessKeys
from aws_iam_login.credentials import Credentials
from aws_iam_login.application_context import ApplicationContext, ApplicationMessages


@click.group(invoke_without_command=True)
@click.option("--debug/--no-debug")
@click.argument("profile")
@click.pass_context
def main(ctx: Context, debug: bool, profile: str) -> None:
    """
    aws-iam-login
    """
    ctx.obj = ApplicationContext(profile=profile, debug=debug)

    if ctx.invoked_subcommand is None:
        credentials(ctx=ctx.obj)


def credentials(ctx: ApplicationContext) -> None:
    config = AWSConfig(ctx.profile)
    session = boto3.Session(profile_name=ctx.profile)
    client = session.client("sts")

    try:
        response = client.get_session_token(
            SerialNumber=config.mfa_serial,
            TokenCode=input(f"Enter MFA code for {config.mfa_serial}: "),
        )

        config.write(
            f"{ctx.profile}-sts", TempCredentials(response.get("Credentials", {}))
        )
        click.echo(f"Login credentials stored in the {ctx.profile}-sts profile!")
    except ClientError as exc:
        click.echo(f"ClientError: {exc}")
        exit(1)
    except ParamValidationError as exc:
        click.echo(f"ValidationError: {exc}")
        exit(1)


@main.command()
@click.pass_obj
def init(ctx: ApplicationContext) -> None:
    """
    Initialize your `.aws/configuration`
    """
    click.echo(f"Looking up additional required data...")

    action = InitializeConfiguration(profile=ctx.profile)
    action.subscribe_subject(ctx.subject)

    if not action.execute():
        click.echo(f"Failed to initialize the {ctx.profile} profile")
        exit(1)

    click.echo(f"The {ctx.profile} profile has been successfully initialized!")


@main.command()
@click.pass_obj
def rotate(ctx: ApplicationContext) -> None:
    """
    Rotate your IAM User credentials
    """
    click.echo(f"Key rotation process in progress")

    action = RotateAccessKeys(profile=ctx.profile)
    action.subscribe_subject(ctx.subject)

    if not action.execute():
        click.echo(f"Failed to rotate the access keys for the {ctx.profile} profile")
        exit(1)

    click.echo(f"Keys successfully rotated for the {ctx.profile} profile")


if __name__ == "__main__":
    main(obj={})



from typing import Optional
import boto3

from aws_iam_login.actions.action import Action
from aws_iam_login.aws_config import AWSConfig


class InitializeConfiguration(Action):
    """
    Understands how to configure the AWS profiles to be used by aws-iam-login.
    """

    __cached_client: Optional[boto3.Session.client] = None

    def __init__(self, profile: str) -> None:
        self.__profile = profile

        self.__config = AWSConfig(profile)
        self.__client = boto3.Session(profile_name=profile).client("sts")

    def execute(self) -> bool:
        if self.__config.mfa_serial and self.__config.username:
            self.info(f"The {self.__profile} profile is already initialized.")
            return True

        try:
            data = self.__client.get_caller_identity()
            mfa_serial = data["Arn"].replace(":user/", ":mfa/")
            username = mfa_serial.split("/")[-1]
            self.info(f"Determined MFA ARN: {mfa_serial}")
            self.info(f"Determined IAM Username: {username}")
            self.__config.initialize(username=username, mfa_serial=mfa_serial)

            return True
        except Exception as exc:
            message = (
                f"Failed to get the caller identity for the {self.__profile} profile."
            )
            self.exception(message=message, exception=exc)

        return False



from typing import Optional, List, Dict


class Credentials:
    def __init__(self, data: Optional[dict]) -> None:
        sanitized_data = self.__sanitize_input(data)
        self._credentials = sanitized_data if sanitized_data else {}

    def __sanitize_input(self, data: Optional[dict]) -> Optional[dict]:
        if not isinstance(data, dict):
            return None

        sanitized = {}

        for key in self._required_fields:
            if key not in data:
                return None

            sanitized[key] = data[key]

        return sanitized

    @property
    def _required_fields(self) -> List[str]:
        return ["AccessKeyId", "SecretAccessKey"]

    @property
    def valid(self) -> bool:
        return bool(self._credentials)

    def __iter__(self):
        yield "aws_access_key_id", self.aws_access_key_id
        yield "aws_secret_access_key", self.aws_secret_access_key

    @property
    def aws_access_key_id(self) -> str:
        return self._credentials.get("AccessKeyId", "")

    @property
    def aws_secret_access_key(self) -> str:
        return self._credentials.get("SecretAccessKey", "")

    @aws_secret_access_key.setter
    def aws_secret_access_key(self, value: str):
        self._credentials["SecretAccessKey"] = value



from typing import Optional, List
import time
import boto3
from functools import lru_cache
from aws_iam_login import AWSConfig
from aws_iam_login.actions.action import Action
from aws_iam_login.access_key import AccessKey


class RotateAccessKeys(Action):
    """
    Understands how to rotate access keys
    """

    __cached_client: Optional[boto3.Session.client] = None

    def __init__(self, profile: str) -> None:
        self.__profile = profile
        self.__config = AWSConfig(profile)

    @lru_cache()
    def new_key(self) -> Optional[AccessKey]:
        try:
            response = self.__client.create_access_key(UserName=self.__config.username)
            self.info(
                message="Sleep for 10 seconds to make sure the credentials are active."
            )
            time.sleep(10)
            return AccessKey(response["AccessKey"])
        except Exception as exc:
            message = f"Failed to create a new access key for the user: {self.__config.username}"
            self.exception(message=message, exception=exc)

        return None

    def execute(self) -> bool:
        if not self.__rotation_possible():
            return False

        new_key = self.new_key()

        if self.__config.key and new_key:
            try:
                self.__flush_client()
                self.__disable_key(self.__config.key)
                self.__config.write(f"{self.__profile}", new_key.credentials)
                self.__delete_key(self.__config.key)

                return True
            except Exception as exc:
                message = f"Failed to rotate the credentials for the user: {self.__config.username}"
                self.exception(message=message, exception=exc)

        return False

    def __flush_client(self) -> None:
        self.__cached_client = None

    @property
    def __client(self) -> boto3.Session.client:
        if not self.__cached_client:
            session = boto3.Session(profile_name=self.__profile)
            self.__cached_client = session.client("iam")

        return self.__cached_client

    def __get_current_keys(self) -> List[dict]:
        try:
            response = self.__client.list_access_keys(UserName=self.__config.username)
            return response["AccessKeyMetadata"]
        except Exception as exc:
            message = (
                f"Failed to list access keys for the user: {self.__config.username}"
            )
            self.exception(message=message, exception=exc)

        return []

    def __rotation_possible(self) -> bool:
        if not self.__config.valid:
            self.warning(
                f"The configuration for the {self.__profile} is invalid, please try `aws-iam-login {self.__profile} init` first!"
            )
            return False

        if len(self.__get_current_keys()) != 1:
            self.error(f"There needs to be only 1 AccessKey present!")
            return False

        return True

    def __disable_key(self, key: AccessKey) -> None:
        self.__client.update_access_key(
            UserName=self.__config.username,
            AccessKeyId=key.credentials.aws_access_key_id,
            Status="Inactive",
        )

    def __delete_key(self, key: AccessKey) -> None:
        self.__client.delete_access_key(
            UserName=self.__config.username,
            AccessKeyId=key.credentials.aws_access_key_id,
        )



from datetime import datetime
from typing import Dict, Union, Optional

from aws_iam_login.credentials import Credentials


class AccessKey:
    """
    Understands the AccessKey format of AWS
    """

    def __init__(self, data: Optional[dict]):
        sanitized_data = self.__sanitize_input(data)
        self.__data = sanitized_data if sanitized_data else {}
        self.__credentials = Credentials(data)

    @staticmethod
    def __sanitize_input(data: Optional[dict]) -> Optional[dict]:
        if not isinstance(data, dict):
            return None

        sanitized = {}

        for key in ["UserName", "Status", "CreateDate"]:
            if key in data:
                sanitized[key] = data[key]

        return sanitized

    @property
    def username(self) -> str:
        return str(self.__data.get("UserName", ""))

    @property
    def credentials(self) -> Credentials:
        return self.__credentials

    @property
    def status(self) -> str:
        return str(self.__data.get("Status", ""))

    @property
    def created(self) -> datetime:
        date = self.__data.get("CreateDate")
        return date if isinstance(date, datetime) else datetime.now()

    def __str__(self) -> str:
        postfix = ""
        if self.credentials.aws_secret_access_key:
            postfix = " (contains secret)"
        return f"{self.credentials.aws_access_key_id}, {self.status} {self.username} created at {self.created}{postfix}"

    def __repr__(self) -> str:
        return str(self)



from typing import Optional, List

from aws_iam_login.credentials import Credentials


class TempCredentials(Credentials):
    def __init__(self, credentials: Optional[dict]) -> None:
        self._required_fields.extend(["SessionToken", "Expiration"])
        super().__init__(credentials)

    def __iter__(self):
        yield "aws_access_key_id", self.aws_access_key_id
        yield "aws_secret_access_key", self.aws_secret_access_key
        yield "aws_session_token", self.aws_session_token
        yield "expiration", self.expiration

    @property
    def _required_fields(self) -> List[str]:
        return super()._required_fields + ["SessionToken", "Expiration"]

    @property
    def aws_session_token(self) -> str:
        return self._credentials.get("SessionToken", "")

    @property
    def expiration(self) -> str:
        return str(self._credentials.get("Expiration", ""))
